#!/usr/bin/env python3

# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)


import kaldifst


# see also https://www.openfst.org/twiki/bin/view/FST/MinimizeDoc
def test_minimize_acceptor():
    sym = kaldifst.SymbolTable()
    sym.add_symbol("a", 1)
    sym.add_symbol("b", 2)
    sym.add_symbol("c", 3)
    sym.add_symbol("d", 4)
    sym.add_symbol("f", 5)
    sym.add_symbol("g", 6)

    s = """
        0 1 a 2
        0 1 b 2
        0 1 c 3
        0 2 d 3
        0 2 f 1
        1 3 f 3
        1 3 g 2
        3 1
        2 4 f 5
        2 4 g 4
        4 3
    """
    fst = kaldifst.compile(s, acceptor=True, isymbols=sym)
    s1 = kaldifst.draw(fst, acceptor=True, isymbols=sym, portrait=True)
    print(s1)  # see https://git.io/JSwRL

    kaldifst.minimize(fst)
    s2 = kaldifst.draw(fst, acceptor=True, isymbols=sym, portrait=True)
    print(s2)  # see https://git.io/JSwRY


# see also https://www.openfst.org/twiki/bin/view/FST/MinimizeDoc
def test_minimize_transducer():
    isym = kaldifst.SymbolTable()
    isym.add_symbol("a", 1)
    isym.add_symbol("b", 2)
    isym.add_symbol("c", 3)
    isym.add_symbol("d", 4)
    isym.add_symbol("f", 5)
    isym.add_symbol("g", 6)

    osym = kaldifst.SymbolTable()
    osym.add_symbol("a", 1)
    osym.add_symbol("b", 2)
    osym.add_symbol("c", 3)

    s = """
        0 1 a a 2
        0 1 b a 2
        0 1 c a 3
        0 2 d a 3
        0 2 f a 1
        1 3 f b 3
        1 3 g b 2
        3 1
        2 4 f c 5
        2 4 g c 4
        4 3
    """
    fst = kaldifst.compile(s, acceptor=False, isymbols=isym, osymbols=osym)
    s1 = kaldifst.draw(
        fst, acceptor=False, isymbols=isym, osymbols=osym, portrait=True
    )
    print(s1)  # see https://git.io/JSw0u

    kaldifst.minimize(fst)
    s2 = kaldifst.draw(
        fst, acceptor=False, isymbols=isym, osymbols=osym, portrait=True
    )
    print(s2)  # see https://git.io/JSwga


def main():
    test_minimize_acceptor()
    test_minimize_transducer()


if __name__ == "__main__":
    main()



# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)
"""
Convert graphs between k2 and OpenFst.
"""

from typing import Optional

import k2
import torch
from _kaldifst import StdArc, StdVectorFst


def _k2_acceptor_to_openfst(fsa: k2.Fsa) -> StdVectorFst:
    """Convert an acceptor in k2 format to a one in OpenFst format.

    Args:
      fsa:
        An acceptor in k2 format.
    Returns:
      Return an acceptor in OpenFst format.
    """
    fsa = fsa.to("cpu")
    assert len(fsa.shape) == 2, "FsaVec is not supported"
    # k2 uses `max` in tropical semiring
    # while OpenFst uses `min`, so we need to
    # negate `fsa.scores`.
    weight = (-1 * fsa.scores).tolist()
    row_splits = fsa.arcs.row_splits(1)
    row_ids = fsa.arcs.row_ids(1)

    num_states = row_ids.numel()
    arcs_per_state = (row_splits[1:] - row_splits[:-1]).tolist()

    fst = StdVectorFst()

    fst.reserve_states(num_states)
    for i in range(num_states):
        s = fst.add_state()
        assert s == i
        fst.reserve_arcs(i, arcs_per_state[i])

    for i, (src, dst, label) in enumerate(fsa.arcs.values()[:, :-1]):
        if label == -1:
            label = 0
        arc = StdArc(
            ilabel=label, olabel=label, weight=weight[i], nextstate=dst
        )
        fst.add_arc(src, arc)

    fst.start = 0
    fst.set_final(num_states - 1, 0)

    return fst


def _k2_transducer_to_openfst(fsa: k2.Fsa, olabels: str) -> StdVectorFst:
    """Convert a transducer in k2 format to an FST in OpenFst format

    Args:
      fsa:
        The input FSA.
      olabels:
        It specifies which attribute from `fsa` contains output labels.
        In most cases, `olabels` is `aux_labels`.
    Returns:
      Return a transducer in OpenFst format.
    """
    assert hasattr(fsa, olabels)
    if isinstance(getattr(fsa, olabels), k2.RaggedTensor):
        fsa = k2.expand_ragged_attributes(fsa, ragged_attribute_names=[olabels])

    fsa = fsa.to("cpu")
    assert len(fsa.shape) == 2, "FsaVec is not supported"

    # k2 uses `max` in tropical semiring
    # while OpenFst uses `min`, so we need to
    # negate `fsa.scores`.
    weight = (-1 * fsa.scores).tolist()
    row_splits = fsa.arcs.row_splits(1)
    row_ids = fsa.arcs.row_ids(1)

    output_labels = getattr(fsa, olabels)
    assert isinstance(output_labels, torch.Tensor)
    output_labels = output_labels.tolist()

    num_states = row_ids.numel()
    arcs_per_state = (row_splits[1:] - row_splits[:-1]).tolist()

    fst = StdVectorFst()
    fst.reserve_states(num_states)

    for i in range(num_states):
        s = fst.add_state()
        assert s == i
        fst.reserve_arcs(i, arcs_per_state[i])

    for i, (src, dst, ilabel) in enumerate(fsa.arcs.values()[:, :-1]):
        olabel = output_labels[i]
        if ilabel == -1:
            ilabel = 0
            assert olabel == -1, olabel
            olabel = 0

        arc = StdArc(
            ilabel=ilabel,
            olabel=olabel,
            weight=weight[i],
            nextstate=dst,
        )
        fst.add_arc(src, arc)

    fst.start = 0
    fst.set_final(num_states - 1, 0)

    return fst


def k2_to_openfst(fsa: k2.Fsa, olabels: Optional[str] = None) -> StdVectorFst:
    """Convert an FSA in k2 format to an FSA in OpenFst format.

    Args:
      fsa:
        An FSA from k2.
        Note:
          Both acceptors and transducers are represented as `k2.Fsa`.
      olabels:
        If it is `None`, we assume the input `fsa` is an acceptor.
        If not `None`, we assume `getattr(fsa, olabel)` contains
        the output labels of arcs. In general, it is calles `aux_labels`
        in k2.
    Return:
     Return an FST in OpenFst format.

    TODO: Also attach symbols, if there are any.
    """
    if olabels is None:
        return _k2_acceptor_to_openfst(fsa)
    else:
        return _k2_transducer_to_openfst(fsa, olabels)



from _kaldifst import (
    FloatWeight,
    StdArc,
    StdConstFst,
    StdVectorFst,
    SymbolTable,
    TropicalWeight,
    arcsort,
    compile,
    compose,
    determinize,
    determinize_star,
    divide,
    draw,
    info,
    invert,
    minimize,
    minimize_encoded,
    plus,
    reverse,
    rmepsilon,
    times,
)

from .iterator import ArcIterator, StateIterator
__version__ = '0.4.1'



#!/usr/bin/env python3

# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)


import kaldifst


# see also https://www.openfst.org/twiki/bin/view/FST/DeterminizeDoc
def test_determinize():
    isym = kaldifst.SymbolTable()
    isym.add_symbol("<eps>", 0)
    isym.add_symbol("a", 1)
    isym.add_symbol("c", 2)
    isym.add_symbol("d", 3)

    osym = kaldifst.SymbolTable()
    osym.add_symbol("<eps>", 0)
    osym.add_symbol("p", 1)
    osym.add_symbol("q", 2)
    osym.add_symbol("r", 3)
    osym.add_symbol("s", 4)

    s = """
        0 1 a p 1
        0 2 a q 2
        1 3 c r 5
        1 3 c r 4
        2 3 d s 6
        3 0
    """

    fst = kaldifst.compile(s, acceptor=False, isymbols=isym, osymbols=osym)
    s1 = kaldifst.draw(
        fst, acceptor=False, isymbols=isym, osymbols=osym, portrait=True
    )
    print(s1)  # see https://git.io/JSPvH

    det_fst = kaldifst.determinize(fst)
    s2 = kaldifst.draw(
        det_fst, acceptor=False, isymbols=isym, osymbols=osym, portrait=True
    )
    print(s2)  # see https://git.io/JSPTF


def main():
    test_determinize()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)


import kaldifst


def test_minimize_encoded_1():
    s = """
        0 0 0 0
        0 0
    """
    fst = kaldifst.compile(s, acceptor=False)
    print(fst)

    kaldifst.minimize_encoded(fst)
    print(fst)


def test_minimize_encoded_2():
    s = """
        0 1 0 0
        0 2 0 0
        1 0
        2 0
    """
    fst = kaldifst.compile(s, acceptor=False)
    print(fst)

    kaldifst.minimize_encoded(fst)
    print(fst)


def main():
    test_minimize_encoded_1()
    test_minimize_encoded_2()


if __name__ == "__main__":
    main()
